<?php

// Exception_DefaultHandler.php - functions to assist when handling errors.
// Based originally on the PHPOF XError.class.php by Tim Jackson
//
// (c)2001-2017 Tim Jackson (tim@timj.co.uk) and contributors
//
// This is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this software.  If not, see <http://www.gnu.org/licenses/>.

class Exception_DefaultHandler
{

	const VERSION = '@PACKAGE-VERSION@';

	const SEND_EMAIL_AUTO = 'auto';
	
	const REDACT_REPLACEMENT = '*********';

	/**
	 * Named variables to redact the values of, in error dumps.
	 *
	 * @see addKeyToRedact
	 * @var array
	 */
	protected static $redact_keys = array();
	
	/**
	 * Known values to redact from error dumps. This might include for example
	 * passwords
	 *
	 * @see addValueToRedact
	 * @var array
	 */
	protected static $redact_values = array();

	/**
	 * Whether to send e-mail error reports to the server administrator
	 * If this is set to Exception_DefaultHandler::SEND_EMAIL_AUTO then
	 * whether or not to send will be decided on the basis of the PHP config
	 * option display_errors, on the basis that if display_errors is on, the
	 * site is under development and doesn't need e-mail error reports. If
	 * display_errors is off, error reports will be sent.
	 */
	public static $send_email = Exception_DefaultHandler::SEND_EMAIL_AUTO;

	public static $email_dest = null;	// will default to $_SERVER['SERVER_ADMIN'];
	
	/**
	 * The e-mail address to be used in the From: header and envelope sender of
	 * e-mailed error reports. If set to NULL, will default to use the same as
	 * $email_dest
	 * 
	 * @var string
	 */
	public static $email_sender = null;

	/**
	 * @var The maximum size of e-mail body to send. Longer mails will be truncated.
	 */
	protected static $max_email_size = 1024000;

	public static $error_page_source = null;
	
	/**
	 * Additional environment information, supplied to Exception_DefaultHandler
	 * @var array
	 */
	protected static $extra_environment_information = array();
	
	/*
	 termination errors as specified by "fatal errors" see:
	 http://uk3.php.net/manual/en/errorfunc.constants.php
	 
	 E_ERROR | E_PARSE | E_CORE_ERROR | E_COMPILE_ERROR | E_USER_ERROR | E_RECOVERABLE_ERROR;
	 int is hardcoded as cannot do bitwise math in static vars at runtime.
	 */
	public static $termination_error = 4437;

	/**
	 * When dumping variables (e.g. HTTP POST data) into error report e-mails,
	 * how many levels deep to recurse into arrays
	 */
	public static $array_recursion_levels = 3;
	
	/**
	 * The minimum interval (in seconds) between error e-mails being sent.
	 * This is to provide a sanity check throttle. Set to 0 to disable 
	 * throttling.
	 *
	 * @var int
	 */
	protected static $min_email_interval = 30;
	
	/**
	 * The file to use to store metadata about throttling of error report
	 * e-mails. The default matches the use of the "data" file role according to
	 * http://pear.php.net/manual/en/guide.users.concepts.filerole.php
	 *
	 * @var string
	 */
	protected static $throttle_file = '@DATA-DIR@/@PACKAGE-NAME@/throttle';
	
	protected static $output_type = null;

	const OUTPUTFORMAT_HTML = 'html';
	const OUTPUTFORMAT_TEXT = 'text';

	/**
	 * Handle exceptions
	 * To use this, set exception handling as follows:
	 *
	 * set_exception_handler(array('Exception_DefaultHandler','handleException'));
	 *
	 * This function is not intended to be called manually; if you want to do
	 * that, use handleExceptionManual().
	 *
	 * @param Exception|Error $exception The exception to handle
	 */
	// In PHP < 7, $exception should be an Exception
	// In PHP >= 7, $exceptions should be an implementation of a Throwable
	// Since this class is not meant to be called manually, we don't explicitly check
	public static function handleException($exception)
	{
		$error_hash = self::getExceptionHash($exception);
		$backtrace = self::getBacktraceOutput($exception->getTrace());
		$class = get_class($exception);
		$message = $exception->getMessage();
		if (strlen($message) > 1024) $message = substr($message, 0, 1024) . ' [error truncated by Exception_DefaultHandler]';
		self::writeLog("$class (code " . $exception->getCode(). ')', $message, $exception->getFile(), $exception->getLine(), $backtrace);
		self::sendEmail($backtrace, null, $exception, $error_hash);
		self::showOutput("$class (code " . $exception->getCode(). ') #' . $error_hash, $message, $exception->getFile(), $exception->getLine(), $backtrace, true);
		die();
	}
	
	/**
	 * Manually handle an exception. Use this method if you want to utilise the
	 * logging/e-mailing functions of this class, but keep control of the
	 * program execution so that you can display output in your own way
	 *
	 * @param Exception|Error $exception The exception to handle
	 */
	public static function handleExceptionManual($exception)
	{
		$phpver = explode('.', PHP_VERSION);
		if ($phpver[0] >= 7 && !$exception instanceof Throwable) {
			die('Exception_DefaultHandler::handleExceptionManual() called but not passed a Throwable');
		} else if (!$exception instanceof Exception) {
			die('Exception_DefaultHandler::handleExceptionManual() called but not passed an Exception');
		}
		$error_hash = self::getExceptionHash($exception);
		$backtrace = self::getBacktraceOutput($exception->getTrace());
		$class = get_class($exception);
		$message = $exception->getMessage();
		if (strlen($message) > 1024) $message = substr($message, 0, 1024) . ' [error truncated by Exception_DefaultHandler]';
		self::writeLog("$class (code " . $exception->getCode(). ')', $message, $exception->getFile(), $exception->getLine(), $backtrace);
		self::sendEmail($backtrace, null, $exception, $error_hash);
		return;
	}

	/**
	 * Handle old-style PHP errors raised by trigger_error() etc. To use this,
	 * set error handling as follows:
	 *
	 * set_error_handler(array('Exception_DefaultHandler','handleError'));
	 *
	 * @param string $errno     The internal PHP error number (an E_* constant)
	 * @param string $errmsg    The error message
	 * @param string $filename  The filename that the error occurred in
	 * @param string $linenum   The line number that the error occurred on
	 * @param array  $vars      Pointer to the active symbol table at the point the error occurred
	 */
	public static function handleError ($errno, $errmsg, $filename, $linenum, $vars)
	{
		//Check bitwise against error_reporting value
		if (($errno & error_reporting()) !== 0) {
			$debug_backtrace = debug_backtrace();
		
			// Clear the first element, which refers to the call to this class
			unset($debug_backtrace[0]);
			
			$backtrace = self::getBacktraceOutput($debug_backtrace);
			$errtype = self::getPHPErrorDescription($errno);
			
			$error_hash[] = strtoupper(substr(md5($errno . $errmsg), 0, 4));
			$error_hash[] = self::locationHash($filename, $linenum, $debug_backtrace);
			$error_hash = implode('/', $error_hash);
			if (strlen($errmsg) > 1024) $errmsg = substr($errmsg, 0, 1024) . ' [error truncated by Exception_DefaultHandler]';
			
			$additional_environment  = "PHP Error Type:     $errtype\n";
			$additional_environment .= "PHP Error Details:  $errmsg\n";
			$additional_environment .= "Script location:    $filename\n";
			$additional_environment .= "Line Number:        $linenum\n";
			self::writeLog($errtype, $errmsg, $filename, $linenum, $backtrace);
			self::sendEmail($backtrace, $additional_environment, $errtype, $error_hash);
			
			// Decide whether this is an error that should result in the
			// termination of execution
			$termination_error = false;
			if (($errno & self::$termination_error) !== 0) {
				$termination_error = true;
			}
			
			self::showOutput($errtype.' #'.$error_hash, $errmsg, $filename, $linenum, $backtrace, $termination_error);
			
			if ($termination_error) {
				die();
			}
		}
	}
	
	/**
	 * Handle "shutdown errors" (errors that are not passed to the normal
	 * defined error handler). Needs PHP 5.2+.
	 * To use this, configure shutdown error handling as follows:
	 *
	 * 	register_shutdown_function(array('Exception_DefaultHandler','handleShutdown'));
	 *
	 * This function will process the last error that could not be handled.
	 * In this "last resort" case, we want to avoid users being shown a blank
	 * page, so show a message when display_errors is turned off (unless errors
	 * should be suppressed for protocol-related reasons)
	 * We don't produce our own error output if display_errors is "on", since
	 * PHP will already have produced error output, there is no clean way of
	 * suppressing it (output buffering cannot be used as when an error occurs
	 * it will be automatically flushed) and there is no point in duplicating it
	 * 
	 * The debug backtrace is not available in the shutdown function. For more
	 * information on that see
	 * http://stackoverflow.com/questions/7241834/debug-backtrace-from-registered-shutdown-function-in-php
	 */
	public static function handleShutdown()
	{
		if (function_exists('error_get_last') && ($error = error_get_last())){
			switch($error['type']){
				case E_ERROR:
				case E_CORE_ERROR:
				case E_COMPILE_ERROR:
				case E_USER_ERROR:
					$debug_backtrace = debug_backtrace();

					// Clear the first element, which refers to the call to this class
					unset($debug_backtrace[0]);

					$backtrace = self::getBacktraceOutput($debug_backtrace);
					$errtype = self::getPHPErrorDescription($error['type']);

					$error_hash[] = strtoupper(substr(md5($error['type'] . $error['message']), 0, 4));
					$error_hash[] = self::locationHash($error['file'], $error['line'], $debug_backtrace);
					$error_hash = implode('/', $error_hash);
					if (strlen($error['message']) > 1024) $error['message'] = substr($error['message'], 0, 1024) . ' [truncated by Exception_DefaultHandler]';

					$additional_environment  = "PHP Error Type:     $errtype\n";
					$additional_environment .= "PHP Error Details:  " . $error['message'] . "\n";
					$additional_environment .= "Script location:    " . $error['file'] . "\n";
					$additional_environment .= "Line Number:        " . $error['line'] . "\n";
					self::writeLog($errtype, $error['message'], $error['file'], $error['line'], $backtrace);
					self::sendEmail($backtrace, $additional_environment, $errtype, $error_hash);

					/*
					 * Only show errors if the errors have not already been shown (display_errors is off)
					 * and errors should not be suppressed.
					 *
					 * Don't use self::getDisplayErrors() here to as it calls self::suppressErrorDisplay() internally
					 * We want to show the error when display errors are turned off and suppress errors is false
					 */
					$disp_err = ini_get('display_errors');
					settype($disp_err, 'string'); // seems to always return string, but be sure
					$disp_err = strtolower(trim($disp_err));
					if ($disp_err !== '1' && $disp_err !== 'on' && !self::suppressErrorDisplay()) {
						// These are always termination errors
						self::showOutput($errtype . ' #' . $error_hash, $error['message'], $error['file'], $error['line'], $backtrace, true);
					}
					break;
				default:
					break;
			}
    	}
	}
	
	/**
	 * Get a unique(ish) hash for a given exception, which merely serves to
	 * tie together related errors when looking at a whole load of errors
	 * The returned hash is in two parts: YYYY/ZZZZ where YYYY is only
	 * dependent on the type of exception; ZZZZ depends on where it is triggered
	 * from
	 *
	 * @param Exception|Error $exception An exception object
	 * @return string              Hash in "YYYY/ZZZZ" format
	 */
	public static function getExceptionHash($exception)
	{
		$phpver = explode('.', PHP_VERSION);
		if ($phpver[0] >= 7 && !$exception instanceof Throwable) {
			die('Exception_DefaultHandler::getExceptionHash() called but not passed a Throwable');
		} else if (!$exception instanceof Exception) {
			die('Exception_DefaultHandler::getExceptionHash() called but not passed an Exception');
		}
		$error_hash[] = strtoupper(substr(md5($exception->getCode() . $exception->getMessage()), 0, 4));
		$error_hash[] = self::locationHash($exception->getFile(), $exception->getLine(), $exception->getTrace());
		$error_hash = implode('/', $error_hash);
		return $error_hash;
	}
	
	/**
	 * Get a unique(ish) partial hash for location of an error or exception.
	 * Hashing uses the file paths in such a way that an the same error in two
	 * different evironments should generate the same hash.
	 *
	 * @param string $filename  The filename that the error occurred in
	 * @param string $linenum   The line number that the error occurred on
	 * @param string $debug_backtrace A backtrace from debug_backtrace() or Exception::getTrace()
	 * @return string              Hash in "YYYYY" format
	 */
	private static function locationHash($filename, $linenum, $debug_backtrace)
	{
		$entropy = basename($filename) . $linenum;
		foreach ($debug_backtrace as $trace_step) {
			if (isset($trace_step['function'])) $entropy .= $trace_step['function'];
			if (isset($trace_step['class']))    $entropy .= $trace_step['class'];
			if (isset($trace_step['type']))     $entropy .= $trace_step['type'];	
		}
		return strtoupper(substr(md5($entropy), 0, 5));
	}
	
	/**
	 * Show details of the error/exception to the output stream, respecting the
	 * value of the display_errors setting and attempting to show either text or
	 * HTML output as appropriate
	 *
	 * @param string $errtype   A description of the error type
	 * @param string $errmsg    The error message
	 * @param string $filename  The filename that the error occurred in
	 * @param string $linenum   The line number that the error occurred on
	 * @param string $backtrace A backtrace from self::getBacktraceOutput()
	 * @param bool   $termination_error  Is the error one that requires the script to terminate?
	 */
	protected static function showOutput($errtype, $errmsg, $filename, $linenum, $backtrace, $termination_error)
	{
		if (self::suppressErrorDisplay()) { // Check for non-interactive requests
			if ($termination_error === true && !headers_sent()) {
				header('HTTP/1.1 500 Internal Server Error');
				header('Cache-control: max-age=0');
				// we return an empty body as there isn't much else we can do; we don't want to send any output as it might be parsed incorrectly
			}
			return;
		}
		
		if (self::getDisplayErrors() == false) {
			/* Generally, we respect the setting of display_errors. However, if
			 * the error is a termination error, we want to show *something* other
			 * than an unhelpful blank page
			 */
			if ($termination_error === true) {
				self::showErrorPage();
			}
			return;
		}
		
		$output = "$errtype: $errmsg in $filename on line $linenum\n\n" . $backtrace;
		if (self::getOutputType() == self::OUTPUTFORMAT_HTML) {
			print $output = "<b>$errtype</b>: " . htmlspecialchars($errmsg, ENT_QUOTES) . " in <b>$filename</b> on line <b>$linenum</b><br />\n";
			print "<pre style='text-align:left;'>" . nl2br(htmlspecialchars($backtrace, ENT_QUOTES)) . '</pre>';
		} else if (PHP_SAPI == 'cli') {
			fwrite(STDERR, $output);
		} else {
			print $output;
		}
	}
	
	/**
	 * Write log in the standard way. How this is handled is configured via the
	 * normal PHP settings.
	 *
	 * @param string $errtype   A description of the error type
	 * @param string $errmsg    The error message
	 * @param string $filename  The filename that the error occurred in
	 * @param string $linenum   The line number that the error occurred on
	 * @param string $backtrace A backtrace from self::getBacktraceOutput()
	 */
	protected static function writeLog($errtype, $errmsg, $filename, $linenum, $backtrace)
	{
		error_log("$errtype: $errmsg in $filename on line $linenum");
	}
	
	protected static function showErrorPage()
	{
		// Clear the output buffer so that only our error page gets sent
		@ob_end_clean();
	
		if (!headers_sent()) {
			header('HTTP/1.1 500 Internal Server Error');
			header('Cache-control: max-age=0');
		}
		
		$source = null;
		if (self::$error_page_source != null) {
			$source = @file_get_contents(self::$error_page_source, true);
		}
		if ($source == null) {
			$source = self::getInternalErrorPage();
		}
		if (self::getOutputType() == self::OUTPUTFORMAT_HTML) {
			print $source;
		} else {
			print strip_tags($source);
		}
		print "\n";
	}
	
	protected static function getInternalErrorPage()
	{
		$output = "<h1>Internal Error</h1>\n";
		$output .= "<p>Sorry, an internal error occurred.";
		if (isset($_SERVER['SERVER_ADMIN'])) {
			$output .= ' Please contact <a href="mailto:'. $_SERVER['SERVER_ADMIN'] . '">' . $_SERVER['SERVER_ADMIN'] . "</a> for more details.\n";
		}
		$output .= "</p>";
		return $output;
	}
	
	protected static function getOutputType()
	{
		if (self::$output_type !== null) {
			return self::$output_type;
		}
		if (PHP_SAPI == 'cli') {
			self::$output_type = self::OUTPUTFORMAT_TEXT;
			return self::$output_type;
		} elseif (isset($_SERVER['HTTP_ACCEPT'])){
			if (preg_match('@text/html|application/xhtml\+xml|\*/\*@',$_SERVER['HTTP_ACCEPT'])) {
				self::$output_type = self::OUTPUTFORMAT_HTML;
				return self::$output_type;
			} else {
				self::$output_type = self::OUTPUTFORMAT_TEXT;
				return self::$output_type;
			}
		} else {
			$headers = headers_list();
			foreach ($headers as $header) {
				if (preg_match('/^Content-type:\s*(.+)$/i', $header, $matches)) {
					if (substr(strtolower($matches[1]), 0, 9) == 'text/html') {
						self::$output_type = self::OUTPUTFORMAT_HTML;
						return self::$output_type;
					} else {
						self::$output_type = self::OUTPUTFORMAT_TEXT;
						return self::$output_type;
					}
				}
			}
		}
		
		// Fallback
		self::$output_type = self::OUTPUTFORMAT_HTML;
		return self::$output_type;
	}

	/*
	* @param array $backtrace A backtrace array from Exception::getTrace or debug_get_backtrace()
	*/
	protected static function getBacktraceOutput($backtrace)
	{
		$level = 0;
		
		$output = 'Backtrace not available';
		if (count($backtrace) > 0) {
			$output = '';
		}
		
		foreach ($backtrace as $trace_step) {
			$prefix_header = str_repeat('-', $level * 2);
			$prefix = str_repeat(' ', $level * 2);

			if (isset($trace_step['file'])) $output .= $prefix_header . 'Called from ' . $trace_step['file'] . "\n";
			if (isset($trace_step['line'])) $output .= $prefix." > Line:     ".$trace_step['line']."\n";
			if (isset($trace_step['type'])) {
				$output .= $prefix . ' > Function: ' . $trace_step['class'] . $trace_step['type'] . $trace_step['function'] . "()\n";
			} else {
				if (isset($trace_step['function'])) $output .= $prefix . ' > Function: ' . $trace_step['function'] . "\n";
				if (isset($trace_step['class']))    $output .= $prefix . ' > Class:    ' . $trace_step['class'] . "\n";
				if (isset($trace_step['type']))     $output .= $prefix . ' > Type:     ' . $trace_step['type'] . "\n";
			}
			
			// Output the arguments passed to a function in the trace
			if (isset($trace_step['args'])) {
				foreach ($trace_step['args'] as $num => $arg)
				{
					$output .= $prefix . str_pad("  > Arg[$num]:", 13);
					
					// We don't use gettype() here as PHP manual recommends against it
					
					// if is scalar, we should be able to output directly
					if (is_scalar($arg)) {
						$type = gettype($arg);
						if (strlen($arg) > 1024) { // Sanity truncation, in case of binary or other large data in arguments
							$arg = substr($arg, 0, 1024) . ' [arg truncated by Exception_DefaultHandler]';
						}
						$output .= $type . "('" . $arg . "')";
					} else if (is_array($arg)) {
						$output .= 'array(';
						$output_array = array();
						foreach ($arg as $key => $val)
						{
							if (in_array($key, self::$redact_keys, true)) {
								$val = self::REDACT_REPLACEMENT;
							}
							if (is_scalar($val)) {
								$output_array[] = "'$key' => '$val'";
							} else {
								$output_array[] = "'$key' => " . gettype($val);
							}
						}
						$output .= implode(', ', $output_array) . ')';
					} else {
						$output .= gettype($arg);
					}
					$output .= "\n";
					
				}
			}
			$level++;
		}
		
		if (count(self::$redact_values) > 0) {
			$output = str_replace(self::$redact_values, self::REDACT_REPLACEMENT, $output);
		}
		
		return $output;
	}
	
	public static function getEnvironment()
	{
		
		$error  = 'Date/Time:          ' . date("D j M Y H:i:s (T)")."\n";
		
		if (isset($_SERVER['HTTP_HOST'])) {

			$request_url = $_SERVER['REQUEST_URI'];
			if (count(self::$redact_keys) > 0) {
				$request_url = preg_replace('@([\?&](?:' . implode('|', array_map('preg_quote', self::$redact_keys)) . ')=)([^\?&]*)@', '$1' . self::REDACT_REPLACEMENT, $request_url);
			}
			
			// Some environments (e.g. FastCGI with mod_fcgid) send the complete
			// URL including scheme in the REQUEST_URI variable
			if (!preg_match('|^[a-z]+://|i', $request_url)) {
				$scheme = 'http';
				if (isset($_SERVER['HTTPS']) && ($_SERVER['HTTPS'] == '1' || $_SERVER['HTTPS'] == 'on')) {
					$scheme = 'https';
				}
				$request_url = $scheme . '://' . $_SERVER['HTTP_HOST'] . $request_url;
			}
			$error .= 'Request URL:        ' . $request_url ."\n";
			if (isset($_SERVER['REMOTE_HOST']) && !empty($_SERVER['REMOTE_HOST'])) {
				$client_host = $_SERVER['REMOTE_HOST'].' ['.$_SERVER['REMOTE_ADDR'].']';
			} else {
				$client_host = $_SERVER['REMOTE_ADDR'];
			}
			$error .= "Client host:        $client_host\n";
			
			if (isset($_ENV['HOSTNAME'])) {
				$error .= 'Server:             ' . $_ENV['HOSTNAME'] . "\n";
			}
			
			// Request method and protocol
			if (isset($_SERVER['REQUEST_METHOD'])) {
				$error .= 'Request method:     ' . $_SERVER['REQUEST_METHOD'];
				if (isset($_SERVER['SERVER_PROTOCOL']) && !empty($_SERVER['SERVER_PROTOCOL'])) {
					$error .= ' (' . $_SERVER['SERVER_PROTOCOL'] . ')';
				}
				$error .= "\n";
			}
			
			// HTTP variables
			$http_vars=null;
			foreach ($_SERVER as $key => $value) {
				if (substr($key,0,5) == 'HTTP_' && $key != 'HTTP_HOST') {
					$http_header = strtolower(substr($key, 5));
					$http_header = ucwords(str_replace('_', ' ', $http_header));
					$http_header = str_replace(' ', '-', $http_header);
					$http_vars .= str_pad($http_header . ':', 20) . "$value\n";
				}
			}
			
			if (isset($http_vars)) {
				$error .= "\nHTTP parameters:\n$http_vars";
			}
			
			if (count($_POST) > 0) {
				$error .= "\nHTTP POST variables:\n";
				$error .= self::getVariableOutput($_POST);
			}
			
			if (isset($_SESSION) && count($_SESSION) > 0) {
				$error .= "\nSession variables:\n";
				$error .= self::getVariableOutput($_SESSION);
			}
		}
		
		if (PHP_SAPI == 'cli') {
			if (isset($_ENV['HOSTNAME'])) {
				$error .= 'Hostname:           ' . $_ENV['HOSTNAME'] . "\n";
			}
			if (isset($GLOBALS['argv'])) {
				$error .= "\nCLI arguments (argv):\n";
				$error .= self::getVariableOutput($GLOBALS['argv']);
			}
		}
		
		if (count(self::$extra_environment_information) > 0) {
			$error .= "\nAdditional environment information (supplied by the application):\n";
			$error .= self::getVariableOutput(self::$extra_environment_information);
		}
		
		if (count(self::$redact_values) > 0) {
			$error = str_replace(self::$redact_values, self::REDACT_REPLACEMENT, $error);
		}
		return $error;
	}
	
	protected static function getVariableOutput($array, $prefix='', $recursion_count=0)
	{
		$out = '';
		foreach ($array as $variable => $value) {
			if (in_array($variable, self::$redact_keys)) {
				$value = self::REDACT_REPLACEMENT;
			}
			$variable = empty($prefix) ? $variable : $prefix . '[' . $variable . ']';
			if (is_array($value)) {
				if ($recursion_count > self::$array_recursion_levels) {
					$out .= $variable . ": [array, recursion terminated]\n";
				} else {
					$recursion_count++;
					$out .= self::getVariableOutput($value, $variable, $recursion_count);
					$recursion_count--;
				}
			} else if (is_scalar($value)) {
				$out .= $variable . ": " . $value . "\n";
			} else {
				$out .= $variable . ": " . gettype($value) . "\n";
			}
		}
		return $out;
	}
	
	protected static function getPHPErrorDescription($errno)
	{
		// Types of error (from http://uk3.php.net/manual/en/errorfunc.constants.php)
    	// In reality not all of these will result in a call to the error handler.
		$errortype = array ();
		$errortypes_conditional = array(
			'E_ERROR' => 'Error',
			'E_WARNING' => 'Warning',
			'E_PARSE' => 'Parse Error',
			'E_NOTICE' => 'Notice',
			'E_CORE_ERROR' => 'Core Error',
			'E_CORE_WARNING' => 'Core Warning',
			'E_COMPILE_ERROR' => 'Compile Error',
			'E_COMPILE_WARNING' => 'Compile Warning',
			'E_USER_ERROR' => 'User Error',
			'E_USER_WARNING' => 'User Warning',
			'E_USER_NOTICE' => 'User Notice',
			'E_STRICT' => 'Strict Notice',
			'E_RECOVERABLE_ERROR' => 'Catchable fatal error',
			'E_DEPRECATED' => 'Deprecated code',
			'E_USER_DEPRECATED' => 'Deprecated code',
		);
		// As some errors/warnings have been added between PHP versions, and we
		// want to allow for the possibility of some being deprecated in future,
		// we check that the constants exist before using them
		foreach ($errortypes_conditional as $type => $name) {
			if (defined($type)) {
				$type_const = constant($type);
				$errortype[$type_const] = $name;
			}
		}
		if (isset($errortype[$errno])) {
			return $errortype[$errno];
		} else {
			return 'Unknown error type (' . $errno . ')';
		}
	}
	
	/**
	 * Send error report e-mail
	 * @param string $backtrace_output       Pre-processed backtrace output (text)
	 * @param string $additional_environment Additional text about the
	 *                                       environment, to put in the e-mail.
	 * @param Exception|string $exception    The exception that triggered the
	 *                                       error, or (if not an exception) a
	 *                                       description of the error type e.g.
	 *                                       "User Type"
	 * @return bool  Whether it was (presumed) successfully sent or not. FALSE indicates a mail was not sent for some reason (not necessarily an error)
	 */
	protected static function sendEmail($backtrace_output, $additional_environment='', $exception=null, $error_hash=null)
	{
		if (self::$send_email == false) {
			return false;
		} else if (self::$send_email === self::SEND_EMAIL_AUTO) {
			// If send_email is set to "auto", we decide on the basis of whether
			// display_errors is on or not.
			if (self::getDisplayErrors() == true) {	
				return false;
			}
		}
		
		$email = self::getEmail($backtrace_output, $additional_environment, $exception, $error_hash);
		if ($email === false) return false;

		// Throttle the mail if necessary
		if (self::$min_email_interval > 0) {
			$last_email = @filemtime(self::$throttle_file);
			$time_since_last_email = time() - $last_email;
			if ($time_since_last_email < self::$min_email_interval) {
				$script_path = getcwd() . DIRECTORY_SEPARATOR . $_SERVER['PHP_SELF'];
				$script_path = str_replace(DIRECTORY_SEPARATOR . DIRECTORY_SEPARATOR, DIRECTORY_SEPARATOR, $script_path); // PHP_SELF has leading slash when called from mod_php but not from CLI
				$error_type = (is_string($exception)) ? $exception : "Exception";
				error_log($script_path . ": not sending e-mail error report for $error_type #$error_hash; last error report e-mail from this system was sent $time_since_last_email seconds ago (throttle=" . self::$min_email_interval . 's)');
				return false;
			}
			@touch(self::$throttle_file);
		}

		// Send the mail
		return mail($email->dest, $email->subject, $email->msg, $email->headers, '-f' . $email->sender);
	}

	/**
	 * Get error report e-mail contents
	 * @param string $backtrace_output       Pre-processed backtrace output (text)
	 * @param string $additional_environment Additional text about the
	 *                                       environment, to put in the e-mail.
	 * @param Exception|string $exception    The exception that triggered the
	 *                                       error, or (if not an exception) a
	 *                                       description of the error type e.g.
	 *                                       "User Type"
	 * @return StdClass|false
	 */
	protected static function getEmail($backtrace_output, $additional_environment='', $exception=null, $error_hash=null)
	{
		$email = new StdClass();
		$email->dest = self::$email_dest;
		$email->subject = '';
		$email->msg = '';
		$email->headers = '';
		$email->sender = self::$email_sender;
		
		// Attempt to figure out where to send the mail to
		if ($email->dest === null) {
			if (isset ($_SERVER) && isset ($_SERVER['SERVER_ADMIN'])) {
				$email->dest = $_SERVER['SERVER_ADMIN'];
			} else if (PHP_SAPI == 'cli') {
				if (isset($_ENV) && isset($_ENV['USERNAME']) && isset($_ENV['HOSTNAME'])) {
					$email->dest = $_ENV['USERNAME'] . '@' . $_ENV['HOSTNAME'];
				}
			}
		}
		
		// If we still have no recipient, we can't do anything
		if ($email->dest === null) {
			if (self::getDisplayErrors() == true) {
				// don't enable this; it triggers setting of a Content-Type
				//print 'Error in ' . __CLASS__ . ": cannot send mail; no recipient specified\n";
			}
			return false;
		}
		
		if ($email->sender === null) {
			$email->sender = $email->dest;
		}

		$error_type = 'Error [Unknown type]';
		if (is_object($exception) && $exception instanceof Exception) {
			$additional_environment  = 'Exception Code:     ' . $exception->getCode() . "\n";
			$additional_environment .= 'Exception Message:  ' . $exception->getMessage() . "\n";
			$additional_environment .= 'Script location:    ' . $exception->getFile() . "\n";
			$additional_environment .= 'Line Number:        ' . $exception->getLine() . "\n";
			$error_type = 'Exception';
		} elseif (is_string($exception)) {
			$error_type = $exception;
			// Strip "User " off the error type description; it's not telling us
			// anything very useful
			if (substr($error_type, 0, 5) == 'User ') {
				$error_type = substr($error_type, 5);
			}
		}
		
		
		$email->msg  = self::getEnvironment() . "\n";
		if (!empty($additional_environment)) {
			$email->msg .= trim($additional_environment) . "\n";
		}
		
		$email->msg .= "\nBacktrace:\n" . $backtrace_output;
		
		// Generate a "site ID" to identify the source of the error; use the HTTP
		// Host: header for HTTP requests, otherwise (e.g. for command line
		// scripts), use the script name
		if (isset($_SERVER['HTTP_HOST'])) {
			$site_id = $_SERVER['HTTP_HOST'];
		} else {
			// Handling of PHP_SELF seems to vary between environments; if PHP_SELF
			// begins with a leading slash, assume that it is an absolute path
			// already, otherwise assume it just contains the script name relative
			// to the working directory
			if ($_SERVER['PHP_SELF']{0} == '/') {
				$site_id = $_SERVER['PHP_SELF'];
			} else {
				$site_id = getcwd() . DIRECTORY_SEPARATOR . $_SERVER['PHP_SELF'];
			}
		}
		
		$email->subject = 'Error report (' . $site_id . '): ' . $error_type;
		if ($error_hash !== null) {
			$email->subject .= ' #' . $error_hash;
		}
		
		$email->headers = 
			"Auto-Submitted: auto-generated\n" .
			"Precedence: bulk\n" . 
			"X-Mailer: Exception_DefaultHandler " . self::VERSION;
		
		// RFC2822 sets a hard line length of 998, so wrap at this
		$email->msg = wordwrap($email->msg, 998, "\n", true);
		
		// Truncate the e-mail if it's too long
		if (strlen($email->msg) > self::$max_email_size) {
			$email->msg = substr($email->msg, 0, self::$max_email_size);
			$email->msg .= "\n\n[Message truncated at " . self::$max_email_size . " bytes due to exceeding max_email_size setting]\n";
		}
		
		return $email;
	}
	
	/**
	 * Set the maximum size of e-mail error reports that will be sent. Longer
	 * mails will be truncated to this length.
	 */
	public static function setMaxEmailSize($size)
	{
		self::$max_email_size = $size;
	}

	protected static function getDisplayErrors()
	{
		if (self::suppressErrorDisplay()) {
			return false;
		}
		
		$disp_err = ini_get('display_errors');
		settype($disp_err, 'string'); // seems to always return string, but be sure
		$disp_err = strtolower(trim($disp_err));
		if ($disp_err === '1' || $disp_err === 'on') {
			return true;
		}

		return false;
	}
	
	/**
	 * Should errors to the output stream be suppressed regardless of the display_errors setting? e.g. due to the request being non-interactive (AJAX, DAV, XMLRPC ...)
	 *
	 * @return bool
	 */
	protected static function suppressErrorDisplay()
	{
		// Check for AJAX requests
		if (isset($_SERVER['HTTP_X_REQUESTED_WITH']) && $_SERVER['HTTP_X_REQUESTED_WITH'] === 'XMLHttpRequest') {
			return true;
		}
		
		// Check for DAV-specific request methods
		if (isset($_SERVER['REQUEST_METHOD'])) {
			$dav_methods = array('PROPFIND', 'PROPPATCH', 'MKCOL', 'DELETE', 'PUT', 'COPY', 'MOVE', 'LOCK', 'UNLOCK');
			if (in_array($_SERVER['REQUEST_METHOD'], $dav_methods)) {
				return true;
			}
		}
		if (isset($_SERVER['HTTP_IF'])) { // RFC4918
			return true;
		}
			
		// Check for SOAP requests
		if (isset($_SERVER['HTTP_SOAPACTION'])) {
			return true;
		}
		if (isset($_SERVER['HTTP_CONTENT_TYPE']) && strstr($_SERVER['HTTP_CONTENT_TYPE'], 'application/soap+xml')) {
			return true;
		}
		
		
		return false;
	}
	
	/**
	 * Add a named array member which will be redacted from error output.
	 * Associative arrays (e.g. HTTP POST, session data) will be checked, and
	 * any array members with keys in this list will have their values redacted.
	 * For example, if you set it to array('password'), then the values of
	 * HTTP POST data fields called "password" will be redacted
	 *
	 * @param string $key Name of array member to redact
	 */
	public function addKeyToRedact($key)
	{
		self::$redact_keys[] = $key;
	}
	
	/**
	 * Add a string value which will be redacted from error output.
	 * This is done by a simple search-and-replace in error output.
	 * For example, if a script is handling a user password, you might add the
	 * password as a value to redact, so that if an error occurs during that
	 * script, the actual password will not appear in the error output.
	 *
	 * @param string $value Value to redact
	 */
	public function addValueToRedact($value)
	{
		self::$redact_values[] = $value;
	}
	
	/**
	 * Add extra environmental information, to be returned with error report
	 * e-mail. Supply in the form of key => value pairs.
	 *
	 * @param array $info Additional information, in key=>value format
	 */
	public static function addEnvironmentInformation(array $info)
	{
		self::$extra_environment_information = array_merge(self::$extra_environment_information, $info);
	}
	
	/**
	 * Return a list of values that have been configured to redact from error reports
	 *
	 * @return array
	 */
	public static function getValuesToRedact()
	{
		return self::$redact_values;
	}

	/**
	 * Return a list of array keys that have been configured to redact from error reports
	 * 
	 * @return array
	 */
	public static function getKeysToRedact()
	{
		return self::$redact_keys;
	}
	
	/**
	 * Reset the throttle timer, to avoid throttling of the next error/exception
	 * 
	 * @return void
	 */
	public static function resetThrottle()
	{
		@touch(self::$throttle_file, 0);
	}
	
}
